/**
 * Copyright 2010 Trustees of the University of Pennsylvania Licensed under the
 * Educational Community License, Version 2.0 (the "License"); you may
 * not use this file except in compliance with the License. You may
 * obtain a copy of the License at
 *
 * http://www.osedu.org/licenses/ECL-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an "AS IS"
 * BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing
 * permissions and limitations under the License.
 */

package metridoc.dsl

import org.springframework.jdbc.datasource.embedded.EmbeddedDatabaseBuilder
import org.springframework.jdbc.datasource.embedded.EmbeddedDatabaseType
import com.mchange.v2.c3p0.ComboPooledDataSource
import org.slf4j.LoggerFactory
import javax.sql.DataSource
import org.apache.camel.CamelContext
import org.apache.camel.spi.BrowsableEndpoint
import org.apache.camel.ShutdownRunningTask
import org.apache.camel.impl.DefaultCamelContext
import org.apache.camel.ProducerTemplate
import metridoc.registry.MetridocSimpleRegistry

import metridoc.utils.ObjectUtils
import liquibase.integration.spring.SpringLiquibase
import metridoc.utils.StringUtils
import org.springframework.core.io.DefaultResourceLoader

/**
 *
 * @author tbarker
 */
@Mixin(PropertyMissing.class)
class JobRunner {

    def CamelContext camelContext
    def ProducerTemplate producerTemplate
    def log = LoggerFactory.getLogger(JobRunner.class)
    def boolean exceptionOccurred = false

    def createSchema(args, Closure closure) {
        def schema = new SchemaBuilder(args)
        schema.createColumns closure
        schema.build()
    }

    def liquibase(args) {
        def liquibase = new SpringLiquibase()
        liquibase.dataSource = args.dataSource
        liquibase.changeLog = args.schema
        liquibase.contexts = StringUtils.EMPTY
        liquibase.resourceLoader = new DefaultResourceLoader()
        liquibase.afterPropertiesSet()
    }

    def run(Closure closure) {
        closure.setDelegate(this)
        closure.setDirective(Closure.DELEGATE_ONLY)
        try {
            closure()
        } catch (Exception ex) {
            exceptionOccurred = true
            log.error "exception occurred while running job", ex
        }
    }

    def runRoute(String route, body=ObjectUtils.NULL, headers=[:]) {
        return getProducerTemplate().requestBodyAndHeaders(route, body, headers);
    }
    
    def runAsyncRoute(String route, body=ObjectUtils.NULL, headers=[:]) {
        return getProducerTemplate().asyncRequestBodyAndHeaders(route, body, headers);
    }

    def runJobs(String... jobs) {
        jobs.each {job ->
            manager.runJob job
        }
    }

    def runRoute(args = null, Closure closure) {
        def routeConfig = new RouteConfig()
        if (args != null) {
            routeConfig = new RouteConfig(args)
        }

        def routeBuilder = new GroovyRouteBuilder(route:closure, jobRunner:this)
        getCamelContext().addRoutes(routeBuilder)

        if (routeConfig.shutdown) {
            Thread.sleep(routeConfig.initialDelayBeforeCompletionCheck)
            def noActivityCount = routeConfig.noActivityCount
            def routes = routeBuilder.routeCollection.routes
            def context = getCamelContext()
            def inflight = context.inflightRepository

            while(noActivityCount > 0) {
                boolean noActivity = true

                routes.each {route ->
                    def fromDefinitions = route.inputs
                    fromDefinitions.each {from ->
                        def uri = from.uri
                        def endpoint = context.getEndpoint(uri)
                        def size = inflight.size(endpoint)

                        if (size > 0) {
                            noActivity = false
                        }

                        if (endpoint instanceof BrowsableEndpoint && noActivity) {
                            size = endpoint.exchanges.size()
                            if (size > 0) {
                                noActivity = false
                            }
                        }

                    }
                }

                if (noActivity) {
                    noActivityCount--
                    log.info "no activity detected, will check for completion ${noActivityCount} more times"
                } else {
                    if (noActivityCount < routeConfig.noActivityCount) {
                        log.info "activity was detected, the no activity counter will be reset from {} to {}", 
                            noActivityCount, routeConfig.noActivityCount
                        noActivityCount = routeConfig.noActivityCount
                    }
                    
                }

                Thread.sleep(routeConfig.completionCheckPeriod)
            }

            routes.each{route ->
                context.stopRoute route
            }
        }

    }

    def CamelContext getCamelContext() {
        if (camelContext == null) {
            def services = manager.services
            def registry = new MetridocSimpleRegistry(services)
            camelContext = new DefaultCamelContext(registry)
            camelContext.shutdownRunningTask = ShutdownRunningTask.CompleteAllTasks
        }

        if (!camelContext.isStarted()) {
            camelContext.start()
        }

        return camelContext
    }

    def ProducerTemplate getProducerTemplate() {
        if (producerTemplate == null) {
            producerTemplate = getCamelContext().createProducerTemplate()
        }

        return producerTemplate
    }

    def DataSource dataSource(args) {
        return new ComboPooledDataSource(args)
    }

    /**
     *creates an embedded in-memory hsql datasource that is generally used for testing
    */
    def DataSource embeddedDataSource() {
        return new EmbeddedDatabaseBuilder().setType(EmbeddedDatabaseType.H2).build()
    }

    def void loadProperties(String... propertyFiles) {
        propertyFiles.each {
            def testFileName = it + "-test.properties"
            def classPathFile = it + ".properties"
            def systemFile = manager.metridocDirectory + "/" + classPathFile
            def boolean fileFound = false

            if (loadPropertyFile(classPathFile)) {
                fileFound = true
            }

            if (loadPropertyFile(systemFile, true)) {
                fileFound = true
            }

            if (loadPropertyFile(testFileName)) {
                fileFound = true
            }

            if (!fileFound) {
                throw new FileNotFoundException("could not find property file ${it}")
            }
        }
    }

    def boolean loadPropertyFile(String propertyFile, boolean systemFile = false) {

        def url
        if (systemFile) {
            def file = new File(propertyFile)

            if (!file.exists()) {
                return false
            }

            url = file.toURI().toURL()
        } else {
            def classLoader = manager.jobClassLoader
            url = classLoader.getResource(propertyFile)
        }

        def boolean foundFile = false

        if (url != null) {

            def props = new Properties()
            props.load(url.newReader())
            def config = new ConfigSlurper().parse(props)
            config.keySet().each {
                def value = config[it]
                manager.services[it] = config[it]
            }
            foundFile = true
        }

        return foundFile
    }


}

